==> oobium_datastore-1.0.0-dev.6.tar.gz: lib/src/persist/feature.dart <==
import 'dart:convert' hide Converter;

import 'package:oobium_datastore/src/convert.dart';
import 'package:oobium_datastore/src/state.dart';
import 'package:oobium_datastore/src/persist/workers.dart';
import 'package:oobium_logging/oobium_logging.dart';

abstract class PersistFeature extends BufferedStateFeature {

  final String path;
  final int version;
  final CompactionStrategy compactionStrategy;
  final Converter converter;
  final OnUpgrade onUpgrade;
  PersistFeature(
    this.path,
    this.version,
    this.compactionStrategy,
    this.converter,
    this.onUpgrade
  );

  final _log = Logger('oobium_datastore').child(StateStore);
  DataWorker? _worker;

  var _recordCount = 0;
  int get recordCount => _recordCount;
  int get modelCount => length;

  void putRecord(DataRecord record);
  void removeRecord(DataRecord record);

  @override
  Future<void> onOpen() async {
    _worker = await DataWorker(path, version, onUpgrade).open();
    await for(final record in _worker!.getData()) {
      if(record.isDelete) {
        removeRecord(record);
      } else {
        putRecord(record);
      }
    }
  }

  @override
  Future<void> onFlush() async {
    await _worker?.flush();
    await _worker?.flush();
  }

  @override
  Future<void> onClose() async {
    await _worker?.flush();
    await _worker?.close();
    _worker = null;
  }

  @override
  Future<void> onDestroy() async {
    await _worker?.destroy();
    _worker = null;
  }

  @override
  void onEvents(StateEvents events) {
    final start = DateTime.now();
    if(events.isNotEmpty) {
      _recordCount += events.length;
      if(_shouldCompact()) {
        compact();
        _log.debug(() => 'batch w/compact executed in: ${DateTime.now().millisecondsSinceEpoch - start.millisecondsSinceEpoch} ms');
      } else {
        _worker?.putData(update: [
          ...events.created(this).map((e) => e.value).map(converter.encodePut),
          ...events.updated(this).map((e) => e.value).map(converter.encodePut),
          ...events.removed(this).map((e) => e.value).map(converter.encodeRemove),
        ]);
        _log.debug(() => 'batch w/out compact executed in: ${DateTime.now().millisecondsSinceEpoch - start.millisecondsSinceEpoch} ms');
        _log.trace(() => '  ops: $events');
      }
    }

    _log.debug(() => 'batch executed in: ${DateTime.now().millisecondsSinceEpoch - start.millisecondsSinceEpoch} ms');
  }

  Future<void> compact() {
    final worker = _worker;
    if(worker != null) {
      _recordCount = length;
      final records = stateValues.map(converter.encodePut).toList();
      return worker.putData(reset: records);
    } else {
      return Future.value();
    }
  }

  bool _shouldCompact() => compactionStrategy.shouldCompact(length, _recordCount);
}

class DataRecord {

  final String type;
  final String modelId;
  final String updateId;
  final String? data;
  DataRecord._(this.type, this.modelId, this.updateId, this.data);

  DataRecord(this.type, this.modelId, this.updateId,
      [Map<String, dynamic>? json]
  ) : data = (json != null) ? jsonEncode(json) : null;

  /// delete: <type>:<modelId>:<updateId>
  /// normal: <type>:<modelId>:<updateId>{jsonData}
  factory DataRecord.fromLine(String line) {
    final i1 = line.indexOf(':');
    final i2 = line.indexOf(':', i1 + 1);
    final i3 = line.indexOf('{', i2 + 1);
    if(i3 == -1) {
      return DataRecord._(
          line.substring(0, i1),
          line.substring(i1 + 1, i2),
          line.substring(i2 + 1),
          null
      );
    } else {
      return DataRecord._(
        line.substring(0, i1),
        line.substring(i1 + 1, i2),
        line.substring(i2 + 1, i3),
        line.substring(i3),
      );
    }
  }

  DataRecord withData(Map<String, dynamic> data) => DataRecord(
      type, modelId, updateId, data
  );

  bool get isDelete => (data == null);
  bool get isNotDelete => !isDelete;

  String toJson() => toString();

  Map<String, dynamic> toJsonData() => jsonDecode(data ?? '{}');

  @override
  toString() => isDelete
      ? '$type:$modelId:$updateId'
      : '$type:$modelId:$updateId$data';
}

class UpgradeEvent {
  final int oldVersion;
  final int newVersion;
  final Stream<DataRecord> oldData;
  UpgradeEvent(this.oldVersion, this.newVersion, this.oldData);
}

typedef OnUpgrade = Stream<DataRecord> Function(UpgradeEvent event);

final dropOnUpgrade = (UpgradeEvent event) {
  return Stream<DataRecord>.empty();
};

abstract class CompactionStrategy {

  const CompactionStrategy();

  bool shouldCompact(int modelCount, int recordCount);
}

class DefaultCompactionStrategy extends CompactionStrategy {

  const DefaultCompactionStrategy();

  @override
  bool shouldCompact(int itemCount, int recordCount) {
    if(recordCount < 100) return false;
    final diff = recordCount - itemCount;
    final ratio = itemCount / recordCount;
    return (diff > 4 && ratio < .75);
  }
}

==> oobium_datastore-1.1.1.tar.gz: lib/src/persist/feature.dart <==
import 'dart:convert' hide Converter;

import 'package:oobium_datastore/src/convert.dart';
import 'package:oobium_datastore/src/state.dart';
import 'package:oobium_datastore/src/persist/workers.dart';
import 'package:oobium_logging/oobium_logging.dart';

abstract class PersistFeature extends BufferedStateFeature implements StateIdProvider {

  final String path;
  final int version;
  final CompactionStrategy compactionStrategy;
  final Converter converter;
  final OnUpgrade onUpgrade;
  PersistFeature(
    this.path,
    this.version,
    this.compactionStrategy,
    this.converter,
    this.onUpgrade
  );

  late final _log = Logger('oobium_datastore').child(StateStore);
  String? _id;
  String get id => _id ?? (throw Exception('id not available'));
  DataWorker? _worker;

  var _recordCount = 0;
  int get recordCount => _recordCount;
  int get modelCount => length;

  void putRecord(DataRecord record);
  void removeRecord(DataRecord record);

  @override
  Future<void> onOpen() async {
    _worker = await DataWorker(path, version, onUpgrade).open();
    await for(final record in _worker!.getData()) {
      if(record.isDelete) {
        removeRecord(record);
      } else {
        putRecord(record);
      }
    }
  }

  @override
  Future<void> onFlush() async {
    await _worker?.flush();
    await _worker?.flush();
  }

  @override
  Future<void> onClose() async {
    await _worker?.flush();
    await _worker?.close();
    _worker = null;
  }

  @override
  Future<void> onDestroy() async {
    await _worker?.destroy();
    _worker = null;
  }

  @override
  void onEvents(StateEvents events) {
    final start = DateTime.now();
    if(events.isNotEmpty) {
      _recordCount += events.length;
      if(_shouldCompact()) {
        compact();
        _log.debug(() => 'batch w/compact executed in: ${DateTime.now().millisecondsSinceEpoch - start.millisecondsSinceEpoch} ms');
      } else {
        _worker?.putData(update: [
          ...events.created(this).map((e) => e.value).map(converter.encodePut),
          ...events.updated(this).map((e) => e.value).map(converter.encodePut),
          ...events.removed(this).map((e) => e.value).map(converter.encodeRemove),
        ]);
        _log.debug(() => 'batch w/out compact executed in: ${DateTime.now().millisecondsSinceEpoch - start.millisecondsSinceEpoch} ms');
        _log.trace(() => '  ops: $events');
      }
    }

    _log.debug(() => 'batch executed in: ${DateTime.now().millisecondsSinceEpoch - start.millisecondsSinceEpoch} ms');
  }

  Future<void> compact() {
    final worker = _worker;
    if(worker != null) {
      _recordCount = length;
      final records = stateValues.map(converter.encodePut).toList();
      return worker.putData(reset: records);
    } else {
      return Future.value();
    }
  }

  bool _shouldCompact() => compactionStrategy.shouldCompact(length, _recordCount);
}

class DataRecord {

  final String type;
  final String modelId;
  final String updateId;
  final String? data;
  DataRecord._(this.type, this.modelId, this.updateId, this.data);

  DataRecord(this.type, this.modelId, this.updateId,
      [Map<String, dynamic>? json]
  ) : data = (json != null) ? jsonEncode(json) : null;

  /// delete: <type>:<modelId>:<updateId>
  /// normal: <type>:<modelId>:<updateId>{jsonData}
  factory DataRecord.fromLine(String line) {
    final i1 = line.indexOf(':');
    final i2 = line.indexOf(':', i1 + 1);
    final i3 = line.indexOf('{', i2 + 1);
    if(i3 == -1) {
      return DataRecord._(
          line.substring(0, i1),
          line.substring(i1 + 1, i2),
          line.substring(i2 + 1),
          null
      );
    } else {
      return DataRecord._(
        line.substring(0, i1),
        line.substring(i1 + 1, i2),
        line.substring(i2 + 1, i3),
        line.substring(i3),
      );
    }
  }

  DataRecord withData(Map<String, dynamic> data) => DataRecord(
      type, modelId, updateId, data
  );

  bool get isDelete => (data == null);
  bool get isNotDelete => !isDelete;

  String toJson() => toString();

  Map<String, dynamic> toJsonData() => jsonDecode(data ?? '{}');

  @override
  toString() => isDelete
      ? '$type:$modelId:$updateId'
      : '$type:$modelId:$updateId$data';
}

class UpgradeEvent {
  final int oldVersion;
  final int newVersion;
  final Stream<DataRecord> oldData;
  UpgradeEvent(this.oldVersion, this.newVersion, this.oldData);
}

typedef OnUpgrade = Stream<DataRecord> Function(UpgradeEvent event);

final dropOnUpgrade = (UpgradeEvent event) {
  return Stream<DataRecord>.empty();
};

abstract class CompactionStrategy {

  const CompactionStrategy();

  bool shouldCompact(int modelCount, int recordCount);
}

class DefaultCompactionStrategy extends CompactionStrategy {

  const DefaultCompactionStrategy();

  @override
  bool shouldCompact(int itemCount, int recordCount) {
    if(recordCount < 100) return false;
    final diff = recordCount - itemCount;
    final ratio = itemCount / recordCount;
    return (diff > 4 && ratio < .75);
  }
}
